#include <cassert>
#include <cstring>
#include <iostream>
#include <stddef.h>
#include <sys/mman.h>

constexpr size_t HPAGE_SIZE = 2 * 1024 * 1024;
constexpr size_t PAGE_SIZE = 4096;

/// ======================================================
///
/// Start of helper code to inspect the page table. See
/// https://www.lukas-barth.net/blog/linux-inspect-page-table/
/// for a background.
///
/// If you are interested in how to allocate huge tables, scroll
/// down to the end of this file.
/// ======================================================

size_t page_index(void *ptr) {
  return reinterpret_cast<size_t>(ptr) / PAGE_SIZE;
}

size_t get_pfn(void *ptr) {
  size_t pi = page_index(ptr);
  auto fp = std::fopen("/proc/self/pagemap", "rb");

  // Each entry in pagemap is 64 bits, i.e, 8 bytes.
  std::fseek(fp, pi * 8, SEEK_SET);

  uint64_t page_info = 0;
  auto result = std::fread(&page_info, 8, 1, fp);
  if (result != 1) {
    std::cerr << "Could not read pagemap at entry " << (pi * 8);
  }

  // check if page is present (bit 63). Otherwise, there is no PFN.
  if (page_info & (static_cast<uint64_t>(1) << 63)) {
    uint64_t pfn = page_info & ((static_cast<uint64_t>(1) << 55) - 1);

    if (pfn == 0) {
      std::cout << "Could not get page frame number. Does the process have the "
                   "CAP_SYS_ADMIN capability?\n";
    }

    return static_cast<size_t>(pfn);
  } else {
    std::cerr << "Page not present.\n";
    return 0;
  }
}

size_t get_pflags(void *ptr) {
  size_t pfn = get_pfn(ptr);
  if (pfn == 0) {
    return 0;
  }

  auto fp = std::fopen("/proc/kpageflags", "rb");
  std::fseek(fp, pfn * 8, SEEK_SET);

  uint64_t pflags = 0;
  auto result = std::fread(&pflags, 8, 1, fp);
  if (result != 1) {
    std::cerr << "Could not read pageflags at entry " << (pfn * 8);
  }

  return static_cast<size_t>(pflags);
}

bool is_ehp(void *ptr) {
  uint64_t flags = get_pflags(ptr);

  return (flags & (static_cast<uint64_t>(1) << 17)); // "huge" flag
}

bool is_thp(void *ptr) {
  uint64_t flags = get_pflags(ptr);

  return (flags & (static_cast<uint64_t>(1) << 22)); // "transparent huge"
}

bool is_huge(void *ptr) { return is_ehp(ptr) || is_thp(ptr); }

/// ===================================================
///
/// End of helper code
///
/// ===================================================

void * allocate_unaligned(size_t size) {
  void *mem = mmap(0, size + PAGE_SIZE, (PROT_READ | PROT_WRITE),
                   (MAP_PRIVATE | MAP_ANONYMOUS), 0, 0);
  assert(mem != MAP_FAILED);
  if ((reinterpret_cast<size_t>(mem) % HPAGE_SIZE) == 0) {
    // We randomly got huge-page-aligned memory. Unmap and map again with a
    // shift of one (non-huge!) page.
    std::cout << "First attempt yielded aligned memory. Remapping.\n";
    void *target_addr =
        reinterpret_cast<void *>(reinterpret_cast<size_t>(mem) + PAGE_SIZE);
    munmap(mem, size + PAGE_SIZE);
    mem = mmap(target_addr, size, (PROT_READ | PROT_WRITE),
               (MAP_PRIVATE | MAP_ANONYMOUS), 0, 0);
    assert(mem != MAP_FAILED);
  }

	return mem;
}

int main() {
  auto size = 4 * HPAGE_SIZE;

	void * mem = allocate_unaligned(size);

  madvise(mem, size, MADV_HUGEPAGE);

  // Make sure all pages are present
  memset(mem, 'x', size);

  std::cout << "Start of memory: Is huge? " << is_huge(mem) << "\n";
  std::cout << "Start of memory: Is THP? " << is_thp(mem) << "\n";

  void *nextHPageAlignedPtr =
      reinterpret_cast<void *>(reinterpret_cast<size_t>(mem) + HPAGE_SIZE -
                               reinterpret_cast<size_t>(mem) % HPAGE_SIZE);

  std::cout << "On the next THP aligned address: Is huge? " << is_huge(nextHPageAlignedPtr) << "\n";
  std::cout << "On the next THP aligned address: Is THP? " << is_thp(nextHPageAlignedPtr) << "\n";
}
